#include "stdafx.h"
#include "resource.h"

#include <stdio.h>
#include <stdlib.h>	// for PCHTTP
#include "str.h"
#include "cehttp.h"
#include "MyHttpSock.h"
#include "httpd.h"
#include "MyConf.h"
#include "httphelper.h"

// global var.
SERVER_INFO	g_ServerInfo;

MyConf		g_conf;

#if defined(DEBUG)||defined(_DEBUG)
#define DBGPRINT(hWnd, nID, nStatus, p) \
	{WCHAR buf[256];wsprintf(buf, L"%03u:%s", nStatus, p);SetDlgItemText(hWnd, nID, buf);}
#define DBGLINE(hWnd, nID) \
	{SetDlgItemInt(hWnd, nID, __LINE__, FALSE);}	
#else
#define DBGPRINT(hWnd, nID, nStatus, p) ;
#define DBGLINE(hWnd, nID) ;
#endif

#define THREAD_ACTIVE	TRUE
#define THREAD_NON_ACTIVE	FALSE

//-----------------------------------------------------------------------------
//	@detail
//		Server thread
//	@param
//		THREADPACK
//	@return
//-----------------------------------------------------------------------------
void thListen(LPVOID pvParams)
{
	CRITICAL_SECTION cs;

	LISTEN_THREAD* ptpListenThread = (LISTEN_THREAD*)pvParams;
	CHILD_THREAD* pChildThread = 
		(CHILD_THREAD*)MALLOC( sizeof(CHILD_THREAD) * g_conf.m_nStartServers );
	CHILD_THREAD* pCT;

	if( pChildThread == NULL )
	{
		SetDlgItemText( ptpListenThread->hWnd, IDC_MESSAGE, L"Memory error." );
		ExitThread(0);
	}
	
	static	int	bBeQuick = FALSE;
	CEHTTPLOGAPI fnLogapi;

	// load server settings
	g_ServerInfo.nPort				= g_conf.m_nListenPort;
	g_ServerInfo.pszCgiRootName		= g_conf.m_pszCgiRootName;
	g_ServerInfo.pszDocRootName		= g_conf.m_pszDocRootName;
	g_ServerInfo.pszSoftware		= g_conf.m_pszServerSoftware;
	g_ServerInfo.pszDefaultDocName	= g_conf.m_pszDefaultDocName;
	g_ServerInfo.pszAccessFileName	= g_conf.m_pszAccessFileName;
	g_ServerInfo.pwszLogFile		= malloc_char2wchar(g_conf.m_pszLogFile);
	g_ServerInfo.nLogSize			= g_conf.m_nLogSize;
	g_ServerInfo.nDNSLookup			= g_conf.m_nDNSLookup;

	// server thread start
	ptpListenThread->bValid = TRUE;

	int ii = 0;

	// ------------------------------------------
	// load logapi
	// ------------------------------------------
	HMODULE hLogDLL = NULL;

	{
		// Load dll which make log file.
		WCHAR pwszLogDll[MAX_PATH] = {'\0'};
		GetCurrentDirW( pwszLogDll, MAX_PATH );

		wcscat( pwszLogDll, LOGAPIFILE );
		
		hLogDLL = ::LoadLibrary( pwszLogDll );

		if( hLogDLL != NULL )
		{
			SetDlgItemText( ptpListenThread->hWnd, IDC_MESSAGE, L"Load log dll." );
			fnLogapi = (CEHTTPLOGAPI)GetProcAddress( hLogDLL, FNC_LOG_API );
			
			if( fnLogapi == NULL )
				SetDlgItemText( ptpListenThread->hWnd, IDC_MESSAGE, L"Load log Addr failed." );

		}
		else
		{
			SetDlgItemText( ptpListenThread->hWnd, IDC_MESSAGE, L"Load log ext failed." );
			fnLogapi = NULL;
		}
	}


	// ------------------------------------------
	// Start listen
	// ------------------------------------------
	MySock s;

	if( SOCKET_ERROR == s.Listen(g_ServerInfo.nPort) )
	{
		SetDlgItemText( ptpListenThread->hWnd, IDC_MESSAGE, L"Listen Failed." );
		FREEIF( pChildThread );
		FREEIF( g_ServerInfo.pwszLogFile );
		ptpListenThread->bValid = FALSE;
		ExitThread(0);
	}

	// ------------------------------------------
	// Initialize enviroments for child thread
	// ------------------------------------------
	pCT = pChildThread;

	for( ii = 0; ii < g_conf.m_nStartServers; ii++, pCT++ )
	{
		pCT->bValid =		THREAD_NON_ACTIVE;
		pCT->pbBeQuick =	&bBeQuick;
		pCT->hWnd =			ptpListenThread->hWnd;
		pCT->hThread =		NULL;
		pCT->sock =			INVALID_SOCKET;
		pCT->nMsgID =		ii;
		pCT->fnLogapi =		fnLogapi;
		pCT->pCSection =	&cs;
	}

	// ------------------------------------------
	// Create child threads
	// ------------------------------------------
	pCT = pChildThread;

	for( ii = 0; ii < g_conf.m_nStartServers; ii++, pCT++ )
	{
		DWORD ThreadID;

		pCT->hThread = CreateThread(
			NULL, 0, (LPTHREAD_START_ROUTINE)thAccept, (LPVOID)pCT, CREATE_SUSPENDED, &ThreadID);

		if( pCT->hThread == NULL )
		{
			FREEIF( pChildThread );
			FREEIF( g_ServerInfo.pwszLogFile );
			SetDlgItemText(ptpListenThread->hWnd, IDC_MESSAGE, L"Create thread failed.");
			ExitThread(0);
		}

		SetDlgItemText(ptpListenThread->hWnd, IDC_THREAD0 + ii, L"Ready.");
	}
	
	// ------------------------------------------
	// Do Accept
	// ------------------------------------------
	DWORD dwCount = 0;
	WCHAR wcCount[36];
	bool bProcessClientRequest;

	InitializeCriticalSection( &cs );

	while( ptpListenThread->bValid )
	{
		SOCKET client_socket;
		sockaddr_in saRemoteHost;

		pCT = pChildThread;

		s.SetTimeOut( 3 );


		if( s.IsReadalbe() > 0 )
		{
			client_socket = s.Accept( &saRemoteHost );
			
			if( client_socket == INVALID_SOCKET )
				continue;

			wsprintf( wcCount, L"Reqest:%u", ++dwCount );
			SetDlgItemText(ptpListenThread->hWnd, IDC_MESSAGE, wcCount);
			bProcessClientRequest = FALSE;

			// Find free thread.
			for( ii = 0; ii < g_conf.m_nStartServers; ii++, pCT++ )
			{
				Sleep(0);

				if( pCT->bValid == THREAD_NON_ACTIVE && bProcessClientRequest == FALSE )
				{
					bBeQuick = FALSE;
					bProcessClientRequest = TRUE;

					// Resume child thread.
					pCT->sock = client_socket;
					memcpy( &(pCT->saRemoteHost), &saRemoteHost, sizeof(sockaddr_in) );
					ResumeThread( pCT->hThread );
					//break;
				}
				else // All other thread be THREAD_ACTIVE
				{
					if( pCT->bValid == THREAD_NON_ACTIVE )
						pCT->sock = INVALID_SOCKET;

					ResumeThread( pCT->hThread );
				}

			}

			// No free thread.
			if( FALSE == bProcessClientRequest )
			{
				SetDlgItemText(ptpListenThread->hWnd, IDC_MESSAGE, L"Server busy...");

				MySock sErrorSock( client_socket );
				sErrorSock.SendData( "HTTP/1.0 503 Server busy\r\n\r\n" );
				sErrorSock.Close();
				// ...If child thread on Keep-Alive, stop it.
				bBeQuick = TRUE;
			}
		} // Readable timeout
	}

	// ------------------------------------------
	// Terminate child threads
	// ------------------------------------------
	HANDLE*	phThreads = (HANDLE*)MALLOC( sizeof(HANDLE) * g_conf.m_nStartServers );

	for( ii = 0, pCT = pChildThread; ii < g_conf.m_nStartServers; ii++, pCT++ )
	{
		phThreads[ii] = pCT->hThread;
		pCT->bValid = FALSE;
		bBeQuick = TRUE;
		ResumeThread(pCT->hThread);
	}


	WaitForMultipleObjects( g_conf.m_nStartServers, phThreads, TRUE, 10000 );

	for( ii = 0, pCT = pChildThread; ii < g_conf.m_nStartServers; ii++, pCT++ )
	{
		if( pCT->hThread )
			CloseHandle( pCT->hThread );
	}

	WSACleanup();

	if( hLogDLL )
	{
		FreeLibrary(hLogDLL);
	}

	FREEIF( phThreads );
	FREEIF( pChildThread );
	FREEIF( g_ServerInfo.pwszLogFile );

	ptpListenThread->bValid = FALSE;

	DeleteCriticalSection( &cs );

	ExitThread(1);

}

//-----------------------------------------------------------------------------
//	@detail
//		The client thread invoked listen thread.
//	@param
//		THREADPACK
//	@return
//-----------------------------------------------------------------------------
void thAccept(LPVOID pvParams)
{
	
	// ------------------------------------------
	// Initialize thread
	// ------------------------------------------
	char szConnection[] = "connection";
	char szKeepAlive[] = "keep-alive";
	char szClose[] = "close";
	char szRemoteHost[32];

	char* pszCon = NULL;
	int bConnectionClose = 0;

	CHILD_THREAD* pChildThread = (CHILD_THREAD*)pvParams;
	MyHttpSock* s = new MyHttpSock();

	// Copy the remote host informations.
	s->m_hrHttpReq.psaRemoteHost = &pChildThread->saRemoteHost;
	s->m_pServerInfo = &g_ServerInfo;

	sockaddr_in* psa = s->m_hrHttpReq.psaRemoteHost;
	u_char* chAddr = (u_char*)&(psa->sin_addr.S_un.S_un_b);
	
	// ------------------------------------------
	// Accept loop
	// ------------------------------------------
	while(pChildThread->bValid != FALSE)
	{
		sprintf( szRemoteHost, "%u.%u.%u.%u", chAddr[0], chAddr[1], chAddr[2], chAddr[3] );

		// Set Accept socket.
		s->SetSocket( pChildThread->sock );

		// @RFC2616 sec.8.1 Persistent connection
		// pipeline will not support.

		int nHttpStatus = 0;
		do
		{
			s->SetTimeOut(1);	// 1sec timeout
			// ----------------------------------------------------------------
			// Recive request...
			// ----------------------------------------------------------------
			nHttpStatus = s->RecvRequest(pChildThread->pbBeQuick);

			if( nHttpStatus != 0 )								// 0==socket error
			{
				// ----------------------------------------------------------------
				// Send response
				// ----------------------------------------------------------------
				
				s->SetTimeOut(5);								// 5sec timeout

				pszCon = s->m_hashReqHeader.GetValueI( szConnection );

				if( pszCon == NULL )
				{
					pszCon = szKeepAlive;						// HTTP/1.1 Parmanent connection
				}

				// ----------------------------------------------------------------
				// Keep-Alive check
				// ----------------------------------------------------------------
				
				if(	!g_conf.m_nKeepAlive						// Global setting
					|| ( 0 == STRICMP( pszCon, szClose )	)	// Connection: close
					|| *pChildThread->pbBeQuick	 				// Don't keep alive flag
					|| ( nHttpStatus >= 400 ))					// Must not Keep alive flag
				{
					bConnectionClose = TRUE;
				}
				else
				{
					bConnectionClose = FALSE;
				}

				// Send response to user-agent.
				nHttpStatus = s->SendResponse( nHttpStatus, bConnectionClose );
			}

			if( pChildThread->fnLogapi && nHttpStatus != 0 && g_ServerInfo.pwszLogFile != NULL )
			{
				EnterCriticalSection( pChildThread->pCSection );
				// Add log.
				s->m_hashReqHeader.Set( "remote_addr", szRemoteHost );

				pChildThread->fnLogapi( g_ServerInfo.pwszLogFile, g_ServerInfo.nLogSize, 
					( nHttpStatus == 999 ) ? 200 : nHttpStatus,	&(s->m_hashReqHeader) );

				LeaveCriticalSection( pChildThread->pCSection );

			}

			s->FreeMemory();

		// Should I Keep alive?
		}while( ( 200 <= nHttpStatus && nHttpStatus < 400 )
				&& ( pszCon != NULL )
				&& ( !bConnectionClose ) );

		// ------------------------------------------
		// Socket close
		// Connection: close
		// ------------------------------------------

		s->Close();

		// ------------------------------------------
		// Is terminate signal on?
		// ------------------------------------------
		if( pChildThread->bValid == FALSE )
			break;

		// ------------------------------------------
		// Suspend... resumed by thListen thread
		// ------------------------------------------
		// To avoid deadlock between this thread and parent thread.
		SetThreadPriority( pChildThread->hThread, THREAD_PRIORITY_ABOVE_NORMAL );

		// When this thread doesn't sleep, the parent thread can do nothing.
		// So, I'm just going to sleep.
		pChildThread->bValid = THREAD_NON_ACTIVE;
		SuspendThread( pChildThread->hThread );
		
		// ... sleep ...

		// Here, This thread now wakeuped by a parent thread. Thread priority is set to normal.
		SetThreadPriority( pChildThread->hThread, THREAD_PRIORITY_NORMAL );
	}
	delete s;
	pChildThread->bValid = THREAD_NON_ACTIVE;
	DBGLINE(pChildThread->hWnd, IDC_THREAD0 + pChildThread->nMsgID);
	ExitThread(1);
}

